/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __MPRIS_INTERFACE_H__
#define __MPRIS_INTERFACE_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.mpris.MediaPlayer2 */

#define TYPE_MEDIA_PLAYER2 (media_player2_get_type ())
#define MEDIA_PLAYER2(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), TYPE_MEDIA_PLAYER2, MediaPlayer2))
#define IS_MEDIA_PLAYER2(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), TYPE_MEDIA_PLAYER2))
#define MEDIA_PLAYER2_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), TYPE_MEDIA_PLAYER2, MediaPlayer2Iface))

struct _MediaPlayer2;
typedef struct _MediaPlayer2 MediaPlayer2;
typedef struct _MediaPlayer2Iface MediaPlayer2Iface;

struct _MediaPlayer2Iface
{
  GTypeInterface parent_iface;


  gboolean (*handle_quit) (
    MediaPlayer2 *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_raise) (
    MediaPlayer2 *object,
    GDBusMethodInvocation *invocation);

  gboolean  (*get_can_quit) (MediaPlayer2 *object);

  gboolean  (*get_can_raise) (MediaPlayer2 *object);

  const gchar * (*get_desktop_entry) (MediaPlayer2 *object);

  gboolean  (*get_has_track_list) (MediaPlayer2 *object);

  const gchar * (*get_identity) (MediaPlayer2 *object);

  const gchar *const * (*get_supported_mime_types) (MediaPlayer2 *object);

  const gchar *const * (*get_supported_uri_schemes) (MediaPlayer2 *object);

};

GType media_player2_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *media_player2_interface_info (void);
guint media_player2_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void media_player2_complete_raise (
    MediaPlayer2 *object,
    GDBusMethodInvocation *invocation);

void media_player2_complete_quit (
    MediaPlayer2 *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void media_player2_call_raise (
    MediaPlayer2 *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean media_player2_call_raise_finish (
    MediaPlayer2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean media_player2_call_raise_sync (
    MediaPlayer2 *proxy,
    GCancellable *cancellable,
    GError **error);

void media_player2_call_quit (
    MediaPlayer2 *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean media_player2_call_quit_finish (
    MediaPlayer2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean media_player2_call_quit_sync (
    MediaPlayer2 *proxy,
    GCancellable *cancellable,
    GError **error);



/* D-Bus property accessors: */
gboolean media_player2_get_can_quit (MediaPlayer2 *object);
void media_player2_set_can_quit (MediaPlayer2 *object, gboolean value);

gboolean media_player2_get_can_raise (MediaPlayer2 *object);
void media_player2_set_can_raise (MediaPlayer2 *object, gboolean value);

gboolean media_player2_get_has_track_list (MediaPlayer2 *object);
void media_player2_set_has_track_list (MediaPlayer2 *object, gboolean value);

const gchar *media_player2_get_identity (MediaPlayer2 *object);
gchar *media_player2_dup_identity (MediaPlayer2 *object);
void media_player2_set_identity (MediaPlayer2 *object, const gchar *value);

const gchar *media_player2_get_desktop_entry (MediaPlayer2 *object);
gchar *media_player2_dup_desktop_entry (MediaPlayer2 *object);
void media_player2_set_desktop_entry (MediaPlayer2 *object, const gchar *value);

const gchar *const *media_player2_get_supported_uri_schemes (MediaPlayer2 *object);
gchar **media_player2_dup_supported_uri_schemes (MediaPlayer2 *object);
void media_player2_set_supported_uri_schemes (MediaPlayer2 *object, const gchar *const *value);

const gchar *const *media_player2_get_supported_mime_types (MediaPlayer2 *object);
gchar **media_player2_dup_supported_mime_types (MediaPlayer2 *object);
void media_player2_set_supported_mime_types (MediaPlayer2 *object, const gchar *const *value);


/* ---- */

#define TYPE_MEDIA_PLAYER2_PROXY (media_player2_proxy_get_type ())
#define MEDIA_PLAYER2_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), TYPE_MEDIA_PLAYER2_PROXY, MediaPlayer2Proxy))
#define MEDIA_PLAYER2_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), TYPE_MEDIA_PLAYER2_PROXY, MediaPlayer2ProxyClass))
#define MEDIA_PLAYER2_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), TYPE_MEDIA_PLAYER2_PROXY, MediaPlayer2ProxyClass))
#define IS_MEDIA_PLAYER2_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), TYPE_MEDIA_PLAYER2_PROXY))
#define IS_MEDIA_PLAYER2_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), TYPE_MEDIA_PLAYER2_PROXY))

typedef struct _MediaPlayer2Proxy MediaPlayer2Proxy;
typedef struct _MediaPlayer2ProxyClass MediaPlayer2ProxyClass;
typedef struct _MediaPlayer2ProxyPrivate MediaPlayer2ProxyPrivate;

struct _MediaPlayer2Proxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  MediaPlayer2ProxyPrivate *priv;
};

struct _MediaPlayer2ProxyClass
{
  GDBusProxyClass parent_class;
};

GType media_player2_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (MediaPlayer2Proxy, g_object_unref)
#endif

void media_player2_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
MediaPlayer2 *media_player2_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
MediaPlayer2 *media_player2_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void media_player2_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
MediaPlayer2 *media_player2_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
MediaPlayer2 *media_player2_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define TYPE_MEDIA_PLAYER2_SKELETON (media_player2_skeleton_get_type ())
#define MEDIA_PLAYER2_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), TYPE_MEDIA_PLAYER2_SKELETON, MediaPlayer2Skeleton))
#define MEDIA_PLAYER2_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), TYPE_MEDIA_PLAYER2_SKELETON, MediaPlayer2SkeletonClass))
#define MEDIA_PLAYER2_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), TYPE_MEDIA_PLAYER2_SKELETON, MediaPlayer2SkeletonClass))
#define IS_MEDIA_PLAYER2_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), TYPE_MEDIA_PLAYER2_SKELETON))
#define IS_MEDIA_PLAYER2_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), TYPE_MEDIA_PLAYER2_SKELETON))

typedef struct _MediaPlayer2Skeleton MediaPlayer2Skeleton;
typedef struct _MediaPlayer2SkeletonClass MediaPlayer2SkeletonClass;
typedef struct _MediaPlayer2SkeletonPrivate MediaPlayer2SkeletonPrivate;

struct _MediaPlayer2Skeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  MediaPlayer2SkeletonPrivate *priv;
};

struct _MediaPlayer2SkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType media_player2_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (MediaPlayer2Skeleton, g_object_unref)
#endif

MediaPlayer2 *media_player2_skeleton_new (void);


/* ------------------------------------------------------------------------ */
/* Declarations for org.mpris.MediaPlayer2.Player */

#define TYPE_MEDIA_PLAYER2_PLAYER (media_player2_player_get_type ())
#define MEDIA_PLAYER2_PLAYER(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), TYPE_MEDIA_PLAYER2_PLAYER, MediaPlayer2Player))
#define IS_MEDIA_PLAYER2_PLAYER(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), TYPE_MEDIA_PLAYER2_PLAYER))
#define MEDIA_PLAYER2_PLAYER_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), TYPE_MEDIA_PLAYER2_PLAYER, MediaPlayer2PlayerIface))

struct _MediaPlayer2Player;
typedef struct _MediaPlayer2Player MediaPlayer2Player;
typedef struct _MediaPlayer2PlayerIface MediaPlayer2PlayerIface;

struct _MediaPlayer2PlayerIface
{
  GTypeInterface parent_iface;



  gboolean (*handle_next) (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_open_uri) (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_Uri);

  gboolean (*handle_pause) (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_play) (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_play_pause) (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_previous) (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_seek) (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation,
    gint64 arg_Offset);

  gboolean (*handle_set_position) (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_TrackId,
    gint64 arg_Position);

  gboolean (*handle_set_volume) (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation,
    gdouble arg_volume);

  gboolean (*handle_stop) (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

  gboolean  (*get_can_control) (MediaPlayer2Player *object);

  gboolean  (*get_can_go_next) (MediaPlayer2Player *object);

  gboolean  (*get_can_go_previous) (MediaPlayer2Player *object);

  gboolean  (*get_can_pause) (MediaPlayer2Player *object);

  gboolean  (*get_can_play) (MediaPlayer2Player *object);

  gboolean  (*get_can_seek) (MediaPlayer2Player *object);

  const gchar * (*get_loop_status) (MediaPlayer2Player *object);

  gdouble  (*get_maximum_rate) (MediaPlayer2Player *object);

  GVariant * (*get_metadata) (MediaPlayer2Player *object);

  gdouble  (*get_minimum_rate) (MediaPlayer2Player *object);

  const gchar * (*get_playback_status) (MediaPlayer2Player *object);

  gint64  (*get_position) (MediaPlayer2Player *object);

  gdouble  (*get_rate) (MediaPlayer2Player *object);

  gboolean  (*get_shuffle) (MediaPlayer2Player *object);

  gdouble  (*get_volume) (MediaPlayer2Player *object);

  void (*seeked) (
    MediaPlayer2Player *object,
    gint64 arg_Position);

};

GType media_player2_player_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *media_player2_player_interface_info (void);
guint media_player2_player_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void media_player2_player_complete_next (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

void media_player2_player_complete_previous (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

void media_player2_player_complete_pause (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

void media_player2_player_complete_play_pause (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

void media_player2_player_complete_stop (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

void media_player2_player_complete_play (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

void media_player2_player_complete_seek (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

void media_player2_player_complete_set_position (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

void media_player2_player_complete_open_uri (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);

void media_player2_player_complete_set_volume (
    MediaPlayer2Player *object,
    GDBusMethodInvocation *invocation);



/* D-Bus signal emissions functions: */
void media_player2_player_emit_seeked (
    MediaPlayer2Player *object,
    gint64 arg_Position);



/* D-Bus method calls: */
void media_player2_player_call_next (
    MediaPlayer2Player *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean media_player2_player_call_next_finish (
    MediaPlayer2Player *proxy,
    GAsyncResult *res,
    GError **error);

gboolean media_player2_player_call_next_sync (
    MediaPlayer2Player *proxy,
    GCancellable *cancellable,
    GError **error);

void media_player2_player_call_previous (
    MediaPlayer2Player *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean media_player2_player_call_previous_finish (
    MediaPlayer2Player *proxy,
    GAsyncResult *res,
    GError **error);

gboolean media_player2_player_call_previous_sync (
    MediaPlayer2Player *proxy,
    GCancellable *cancellable,
    GError **error);

void media_player2_player_call_pause (
    MediaPlayer2Player *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean media_player2_player_call_pause_finish (
    MediaPlayer2Player *proxy,
    GAsyncResult *res,
    GError **error);

gboolean media_player2_player_call_pause_sync (
    MediaPlayer2Player *proxy,
    GCancellable *cancellable,
    GError **error);

void media_player2_player_call_play_pause (
    MediaPlayer2Player *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean media_player2_player_call_play_pause_finish (
    MediaPlayer2Player *proxy,
    GAsyncResult *res,
    GError **error);

gboolean media_player2_player_call_play_pause_sync (
    MediaPlayer2Player *proxy,
    GCancellable *cancellable,
    GError **error);

void media_player2_player_call_stop (
    MediaPlayer2Player *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean media_player2_player_call_stop_finish (
    MediaPlayer2Player *proxy,
    GAsyncResult *res,
    GError **error);

gboolean media_player2_player_call_stop_sync (
    MediaPlayer2Player *proxy,
    GCancellable *cancellable,
    GError **error);

void media_player2_player_call_play (
    MediaPlayer2Player *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean media_player2_player_call_play_finish (
    MediaPlayer2Player *proxy,
    GAsyncResult *res,
    GError **error);

gboolean media_player2_player_call_play_sync (
    MediaPlayer2Player *proxy,
    GCancellable *cancellable,
    GError **error);

void media_player2_player_call_seek (
    MediaPlayer2Player *proxy,
    gint64 arg_Offset,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean media_player2_player_call_seek_finish (
    MediaPlayer2Player *proxy,
    GAsyncResult *res,
    GError **error);

gboolean media_player2_player_call_seek_sync (
    MediaPlayer2Player *proxy,
    gint64 arg_Offset,
    GCancellable *cancellable,
    GError **error);

void media_player2_player_call_set_position (
    MediaPlayer2Player *proxy,
    const gchar *arg_TrackId,
    gint64 arg_Position,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean media_player2_player_call_set_position_finish (
    MediaPlayer2Player *proxy,
    GAsyncResult *res,
    GError **error);

gboolean media_player2_player_call_set_position_sync (
    MediaPlayer2Player *proxy,
    const gchar *arg_TrackId,
    gint64 arg_Position,
    GCancellable *cancellable,
    GError **error);

void media_player2_player_call_open_uri (
    MediaPlayer2Player *proxy,
    const gchar *arg_Uri,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean media_player2_player_call_open_uri_finish (
    MediaPlayer2Player *proxy,
    GAsyncResult *res,
    GError **error);

gboolean media_player2_player_call_open_uri_sync (
    MediaPlayer2Player *proxy,
    const gchar *arg_Uri,
    GCancellable *cancellable,
    GError **error);

void media_player2_player_call_set_volume (
    MediaPlayer2Player *proxy,
    gdouble arg_volume,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean media_player2_player_call_set_volume_finish (
    MediaPlayer2Player *proxy,
    GAsyncResult *res,
    GError **error);

gboolean media_player2_player_call_set_volume_sync (
    MediaPlayer2Player *proxy,
    gdouble arg_volume,
    GCancellable *cancellable,
    GError **error);



/* D-Bus property accessors: */
const gchar *media_player2_player_get_playback_status (MediaPlayer2Player *object);
gchar *media_player2_player_dup_playback_status (MediaPlayer2Player *object);
void media_player2_player_set_playback_status (MediaPlayer2Player *object, const gchar *value);

const gchar *media_player2_player_get_loop_status (MediaPlayer2Player *object);
gchar *media_player2_player_dup_loop_status (MediaPlayer2Player *object);
void media_player2_player_set_loop_status (MediaPlayer2Player *object, const gchar *value);

gdouble media_player2_player_get_rate (MediaPlayer2Player *object);
void media_player2_player_set_rate (MediaPlayer2Player *object, gdouble value);

gboolean media_player2_player_get_shuffle (MediaPlayer2Player *object);
void media_player2_player_set_shuffle (MediaPlayer2Player *object, gboolean value);

GVariant *media_player2_player_get_metadata (MediaPlayer2Player *object);
GVariant *media_player2_player_dup_metadata (MediaPlayer2Player *object);
void media_player2_player_set_metadata (MediaPlayer2Player *object, GVariant *value);

gdouble media_player2_player_get_volume (MediaPlayer2Player *object);
void media_player2_player_set_volume (MediaPlayer2Player *object, gdouble value);

gint64 media_player2_player_get_position (MediaPlayer2Player *object);
void media_player2_player_set_position (MediaPlayer2Player *object, gint64 value);

gdouble media_player2_player_get_minimum_rate (MediaPlayer2Player *object);
void media_player2_player_set_minimum_rate (MediaPlayer2Player *object, gdouble value);

gdouble media_player2_player_get_maximum_rate (MediaPlayer2Player *object);
void media_player2_player_set_maximum_rate (MediaPlayer2Player *object, gdouble value);

gboolean media_player2_player_get_can_go_next (MediaPlayer2Player *object);
void media_player2_player_set_can_go_next (MediaPlayer2Player *object, gboolean value);

gboolean media_player2_player_get_can_go_previous (MediaPlayer2Player *object);
void media_player2_player_set_can_go_previous (MediaPlayer2Player *object, gboolean value);

gboolean media_player2_player_get_can_play (MediaPlayer2Player *object);
void media_player2_player_set_can_play (MediaPlayer2Player *object, gboolean value);

gboolean media_player2_player_get_can_pause (MediaPlayer2Player *object);
void media_player2_player_set_can_pause (MediaPlayer2Player *object, gboolean value);

gboolean media_player2_player_get_can_seek (MediaPlayer2Player *object);
void media_player2_player_set_can_seek (MediaPlayer2Player *object, gboolean value);

gboolean media_player2_player_get_can_control (MediaPlayer2Player *object);
void media_player2_player_set_can_control (MediaPlayer2Player *object, gboolean value);


/* ---- */

#define TYPE_MEDIA_PLAYER2_PLAYER_PROXY (media_player2_player_proxy_get_type ())
#define MEDIA_PLAYER2_PLAYER_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), TYPE_MEDIA_PLAYER2_PLAYER_PROXY, MediaPlayer2PlayerProxy))
#define MEDIA_PLAYER2_PLAYER_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), TYPE_MEDIA_PLAYER2_PLAYER_PROXY, MediaPlayer2PlayerProxyClass))
#define MEDIA_PLAYER2_PLAYER_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), TYPE_MEDIA_PLAYER2_PLAYER_PROXY, MediaPlayer2PlayerProxyClass))
#define IS_MEDIA_PLAYER2_PLAYER_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), TYPE_MEDIA_PLAYER2_PLAYER_PROXY))
#define IS_MEDIA_PLAYER2_PLAYER_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), TYPE_MEDIA_PLAYER2_PLAYER_PROXY))

typedef struct _MediaPlayer2PlayerProxy MediaPlayer2PlayerProxy;
typedef struct _MediaPlayer2PlayerProxyClass MediaPlayer2PlayerProxyClass;
typedef struct _MediaPlayer2PlayerProxyPrivate MediaPlayer2PlayerProxyPrivate;

struct _MediaPlayer2PlayerProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  MediaPlayer2PlayerProxyPrivate *priv;
};

struct _MediaPlayer2PlayerProxyClass
{
  GDBusProxyClass parent_class;
};

GType media_player2_player_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (MediaPlayer2PlayerProxy, g_object_unref)
#endif

void media_player2_player_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
MediaPlayer2Player *media_player2_player_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
MediaPlayer2Player *media_player2_player_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void media_player2_player_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
MediaPlayer2Player *media_player2_player_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
MediaPlayer2Player *media_player2_player_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define TYPE_MEDIA_PLAYER2_PLAYER_SKELETON (media_player2_player_skeleton_get_type ())
#define MEDIA_PLAYER2_PLAYER_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), TYPE_MEDIA_PLAYER2_PLAYER_SKELETON, MediaPlayer2PlayerSkeleton))
#define MEDIA_PLAYER2_PLAYER_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), TYPE_MEDIA_PLAYER2_PLAYER_SKELETON, MediaPlayer2PlayerSkeletonClass))
#define MEDIA_PLAYER2_PLAYER_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), TYPE_MEDIA_PLAYER2_PLAYER_SKELETON, MediaPlayer2PlayerSkeletonClass))
#define IS_MEDIA_PLAYER2_PLAYER_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), TYPE_MEDIA_PLAYER2_PLAYER_SKELETON))
#define IS_MEDIA_PLAYER2_PLAYER_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), TYPE_MEDIA_PLAYER2_PLAYER_SKELETON))

typedef struct _MediaPlayer2PlayerSkeleton MediaPlayer2PlayerSkeleton;
typedef struct _MediaPlayer2PlayerSkeletonClass MediaPlayer2PlayerSkeletonClass;
typedef struct _MediaPlayer2PlayerSkeletonPrivate MediaPlayer2PlayerSkeletonPrivate;

struct _MediaPlayer2PlayerSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  MediaPlayer2PlayerSkeletonPrivate *priv;
};

struct _MediaPlayer2PlayerSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType media_player2_player_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (MediaPlayer2PlayerSkeleton, g_object_unref)
#endif

MediaPlayer2Player *media_player2_player_skeleton_new (void);


G_END_DECLS

#endif /* __MPRIS_INTERFACE_H__ */
